/* sgrotum - an email SiGnature ROTUMbulator
 *
 * Copyright (C) 1999-2003, Timothy Musson, trmusson@ihug.co.nz
 *
 * Released under the GNU General Public License:
 * http://www.gnu.org/copyleft/gpl.html
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA  02111-1307, USA.
 *
 */


#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <glib.h>
#include <getopt.h>


/* see configure.in, config.h */
#ifndef PACKAGE
#define PACKAGE "sgrotum"
#endif
#ifndef VERSION
#define VERSION "unknown"
#endif
#ifndef OLDSTYLE
#define OLDSTYLE 0
#endif


/* 0..n-1 */
#define RANDOM_INT(n)         ((int)((double)rand() / ((double)RAND_MAX + 1) * n))

#define PROG_WARN(args...)    g_printerr (PACKAGE ": " args)
#define PROG_WARN_MEM         g_printerr (PACKAGE ": out of memory\n")
#ifdef DEBUG
#define PROG_DEBUG(args...)   g_printerr (PACKAGE ": " args)
#else
#define PROG_DEBUG(args...)
#endif


typedef enum {
        ALIGN_LEFT = 0,
        ALIGN_RIGHT,
        ALIGN_CENTER
} Align;

typedef enum {
        ROTMODE_ROTATE = 0,
        ROTMODE_RANDOM
} RotMode;


typedef struct _Sgr Sgr;
typedef struct _Box Box;

struct _Box {
        gint        x1;             /* left */
        gint        y1;             /* top */
        gint        x2;             /* right */
        gint        y2;             /* bottom */
        gint        minlen;         /* minimum length of quote to insert */
        Align       align;          /* quote alignment within box */
        gboolean    strip;          /* wrap on space and remove leading space? */
        gchar       *env;           /* value of an environment variable */
        gchar       *fname_quotes;  /* name of file containing quotes for this box */
        GList       *quotes;        /* list (gchar*) of quotes that fit x1,y1,x2,y2 */
        gint        id;             /* box number (used in error messages) */
};

struct _Sgr {
        gboolean    dash;           /* prefix result with '-- \n'? */
        GList       *lines;         /* list (GString*) of lines being worked on */
        const gchar *fname_prefs;   /* name of configuration file */
        gchar       *fname_template;/* name of selected signature template file */
        gchar       *fname_quotes;  /* name of default quote file */
        gchar       *fname_dest;    /* name of signature file to write */
        GList       *template_list; /* list (gchar*) of template file names */
        GList       *pref;          /* list (gchar*) of lines in fname_prefs */
        gint        current;        /* currently selected signature template number/id */
        RotMode     rotmode;        /* how to update Sgr.current */
        GList       *box;           /* list (Box*) of quote boxes */
};

static gchar *attrs[] = {
        "x1",
        "y1",
        "x2",
        "y2",
#if OLDSTYLE
        "startx",
        "starty",
        "endx",
        "endy",
#endif
        "align",
        "minlen",
        "strip",
        "file",
        "env"
};
typedef enum {
        ATTR_X1 = 0,
        ATTR_Y1,
        ATTR_X2,
        ATTR_Y2,
#if OLDSTYLE
        ATTR_STARTX,
        ATTR_STARTY,
        ATTR_ENDX,
        ATTR_ENDY,
#endif
        ATTR_ALIGN,
        ATTR_MINLEN,
        ATTR_STRIP,
        ATTR_FILE,
        ATTR_ENV,
        ATTR_ERROR
} AttrType;


static struct option long_options[] = {
        { "fast",     no_argument, NULL, 'f' },
        { "help",     no_argument, NULL, 'h' },
        { "no-write", no_argument, NULL, 'n' },
        { "show",     no_argument, NULL, 's' },
        { "version",  no_argument, NULL, 'v' },
        { 0, 0, 0, 0 }
};


static void cleanup (Sgr *s, gint code);


/* if FALSE, don't write result to target (the --no-write option) */
static gboolean write = TRUE;

/* if TRUE, write result to stdout (the --show option) */
static gboolean show = FALSE;

/* if TRUE, do fast but less accurate quote length checks (the --fast option) */
static gboolean fast = FALSE;



#ifndef __USE_GNU
#define GETLINE_GROW_SIZE 256
static gint
getline (gchar **lineptr, gint *size, FILE *stream)
{
        /*
         * Read a line (till EOF, or up to and including '\n', whichever
         * happens first) from stream into *lineptr.
         * - If *lineptr is NULL, allocate some initial space for it.
         * - Reallocate *lineptr, updating *size, as the string grows.
         * - Return the number of chars read, or -1 if none.
         * The string put into *lineptr will be '\0' terminated.
         */

        gint ch = 0, n = 0;
        gchar *pos;


        if (lineptr == NULL || size == NULL || stream == NULL) return -1;
        if (*lineptr == NULL) {
                *size = GETLINE_GROW_SIZE;
                *lineptr = g_malloc (*size);
                if (*lineptr == NULL) {
                        *size = 0;
                        PROG_WARN_MEM;
                        return -1;
                }
        }

        pos = *lineptr;
        while (ch != '\n' && (ch = fgetc (stream)) != EOF) {
                if (n == *size) {
                        *size = *size + GETLINE_GROW_SIZE;
                        *lineptr = g_realloc (*lineptr, *size);
                        pos = *lineptr + n;
                        if (*lineptr == NULL) {
                                *size = 0;
                                PROG_WARN_MEM;
                                return -1;
                        }
                }

                *pos = ch;

                pos++;
                n++;
        }
        *pos = '\0';

        if (n > 0) return n;
        return -1;
}
#endif



static gchar*
expand_tilde (const gchar *fname)
{
        /*
         * If fname begins with a tilde, return a copy with the
         * tilde replaced by the user's home path. Otherwise
         * return a straight copy of fname.
         */

        gchar *fname_new;
        const gchar *home;

        if (fname == NULL) return NULL;
        if (fname[0] != '~') {
                fname_new = g_strdup (fname);
        }
        else {
                home = g_getenv ("HOME");
                if (home == NULL) {
                        PROG_WARN ("expand_tilde: failed to get HOME\n");
                        return NULL;
                }
                fname_new = g_strdup_printf ("%s%s", home, &fname[1]);
        }
        if (fname_new == NULL) {
                PROG_WARN_MEM;
                exit (1);
        }
        return fname_new;
}



static const gchar*
skip_space (Box *b, const gchar *str)
{
        /*
         * Return a pointer to the next non-space in str,
         * or NULL if no space found before '\0'.
         */

        while (isspace (*str)) str++;
        if (*str == '\0') {
                PROG_WARN ("box %d: unexpected end of string\n", b->id);
                return NULL;
        }
        return str;
}



static const gchar*
skip_chars (Box *b, const gchar *str, gint n)
{
        /*
         * Return &str[n], or NULL if str is too short.
         */

        while (n > 0 && *str != '\0') {str++; n--;}
        if (n != 0) {
                PROG_WARN ("box %d: unexpected end of string\n", b->id);
                return NULL;
        }
        return str;
}



static gboolean
keep (const gchar *str)
{
        /*
         * Return FALSE if str begins with '\0', contains
         * only whitespace, or the first non-space char is '#'.
         */

        while (isspace (*str)) str++;
        return *str != '#' && *str != '\0';
}



static AttrType
identify_attr (const gchar *str, gint *len)
{
        /*
         * Return an AttrType identifying the attribute name
         * beginning at str. Set *len to the length of the
         * attribute, or to 0 on error.
         */

        AttrType attr;

        for (attr = ATTR_X1; attr < ATTR_ERROR; attr++) {
                *len = strlen (attrs[attr]);
                if (g_ascii_strncasecmp (str, attrs[attr], *len) == 0) {
                        if (str[*len] == '=' || isspace (str[*len])) {
                                return attr;
                        }
                }
        }
        *len = 0;
        return ATTR_ERROR;
}



static Box*
box_new (Sgr *s, gint id)
{
        /*
         * Return a newly allocated Box with
         * default settings.
         */

        Box *b;

        b = g_new (Box, 1);
        if (b == NULL) {
                PROG_WARN_MEM;
                exit (1);
        }
        b->x1     = 1;
        b->y1     = 1;
        b->x2     = 70;
        b->y2     = 4;
        b->minlen = 0;
        b->align  = ALIGN_LEFT;
        b->strip  = TRUE;
        b->quotes = NULL;
        b->env    = NULL;
        b->fname_quotes = expand_tilde (s->fname_quotes);

        b->id = id;

        return b;
}



static void
box_free (Box *b)
{
        /*
         * Free this Box and anything
         * allocated for it.
         */

        if (b == NULL) return;
        g_free (b->fname_quotes);
        g_free (b->env);
        if (b->quotes != NULL) {
                g_list_foreach (b->quotes, (GFunc)g_free, NULL);
                g_list_free (b->quotes);
        }
        g_free (b);
}



static Sgr*
sgrotum_new (void)
{
        /*
         * Return a newly allocated Sgr.
         */

        Sgr *s;

        s = g_new0 (Sgr, 1);
        if (s == NULL) {
                PROG_WARN_MEM;
                exit (1);
        }
        s->dash = TRUE;
        s->rotmode = ROTMODE_RANDOM;
        s->current = 1;

        return s;
}



static void
sgrotum_free (Sgr *s)
{
        /*
         * Free s and everything it contains.
         */

        GList *l;


        if (s == NULL) return;

        /* boxes */
        if (s->box != NULL) {
                g_list_foreach (s->box, (GFunc)box_free, NULL);
                g_list_free (s->box);
        }

        /* lines */
        l = s->lines;
        while (l != NULL) {
                g_string_free ((GString*)l->data, TRUE);
                l = l->next;
        }
        g_list_free (s->lines);

        /* prefs */
        if (s->pref != NULL) {
                g_list_foreach (s->pref, (GFunc)g_free, NULL);
                g_list_free (s->pref);
        }

        /* template file names */
        if (s->template_list != NULL) {
                g_list_foreach (s->template_list, (GFunc)g_free, NULL);
                g_list_free (s->template_list);
        }

        g_free (s->fname_quotes);
        g_free (s->fname_template);
        g_free (s->fname_dest);
        g_free (s);
}



static void
cleanup (Sgr *s, gint code)
{
        /*
         * Free all and exit.
         */

        sgrotum_free (s);
        exit (code);
}



static void
write_signature (Sgr *s)
{
        /*
         * Output each line of s->lines
         */

        FILE *f;
        gchar *fname_dest;
        GList *l;


        /* if there's a target file and write is TRUE, write to the target */
        if (write && s->fname_dest != NULL) {

                fname_dest = expand_tilde (s->fname_dest);
                if (fname_dest == NULL) return;

                f = fopen (fname_dest, "w");
                if (f == NULL) {
                        PROG_WARN ("unable to write %s\n", fname_dest);
                        g_free (fname_dest);
                        return;
                }

                if (s->dash) fprintf (f, "-- \n");

                l = s->lines;
                while (l != NULL) {
                        fprintf (f, "%s\n", ((GString*)l->data)->str);
                        l = l->next;
                }

                fclose (f);

                g_free (fname_dest);

        }

        /* if there's no target file or show is TRUE, write to stdout */
        if (show || s->fname_dest == NULL) {
                if (s->dash) fprintf (stdout, "-- \n");
                l = s->lines;
                while (l != NULL) {
                        fprintf (stdout, "%s\n", ((GString*)l->data)->str);
                        l = l->next;
                }
        }
}



static void
write_prefs (Sgr *s)
{
        /*
         * If s->rotmode is ROTMODE_ROTATE, and there's
         * more than one template file, it's necessary
         * to save the value of s->current. That means
         * rewriting the prefs file. Yuck.
         */

        FILE *f;
        GList *l;


        if (s->rotmode == ROTMODE_RANDOM) return;
        if (g_list_length (s->template_list) < 2) return;

        f = fopen (s->fname_prefs, "w");
        if (f == NULL) {
                PROG_WARN ("unable to update %s\n", s->fname_prefs);
                return;
        }

        l = s->pref;
        while (l != NULL) {
                if (g_ascii_strncasecmp ("[current]", (gchar*)l->data, 9) == 0) {
                        break;
                }
                fprintf (f, "%s", (gchar*)l->data);
                l = l->next;
        }

        fprintf (f, "[current]\n%d\n", s->current);

        fclose (f);
}



typedef enum {
        PREFS_NONE = 0,
        PREFS_TARGET,
        PREFS_DASH,
        PREFS_ROTMODE,
#if OLDSTYLE
        PREFS_METHOD,
#endif
        PREFS_QUOTES,
        PREFS_TEMPLATES,
        PREFS_CURRENT
} PrefState;



static gboolean
read_prefs (Sgr *s, const gchar *fname)
{
        /*
         * Read the specified preference file.
         * Return FALSE on failure.
         */

        FILE      *f;
        gint      len, bufsize = 0;
        gchar     **buf;
        PrefState state = PREFS_NONE;


        s->fname_prefs = fname;

        f = fopen (s->fname_prefs, "r");
        if (f == NULL) {
                PROG_WARN ("unable to read %s\n", s->fname_prefs);
                return FALSE;
        }

        buf = g_malloc (sizeof (gchar*));
        if (buf == NULL) {
                fclose (f);
                PROG_WARN_MEM;
                exit (1);
        }
        *buf = NULL;

        do {
                len = getline (buf, &bufsize, f);
                if (len != -1) {
                        /* store line (see write_prefs) */
                        s->pref = g_list_append (s->pref, g_strdup (*buf));
                        if (s->pref == NULL) {
                                fclose (f);
                                PROG_WARN_MEM;
                                exit (1);
                        }
                }
                if (len > 1 && keep (*buf)) {
                        if (buf[0][len - 1] == '\n') buf[0][len - 1] = '\0';
                        if (g_ascii_strcasecmp (*buf, "[target]") == 0) {
                                state = PREFS_TARGET;
                        }
                        else if (g_ascii_strcasecmp (*buf, "[dash]") == 0) {
                                state = PREFS_DASH;
                        }
                        else if (g_ascii_strcasecmp (*buf, "[mode]") == 0) {
                                state = PREFS_ROTMODE;
                        }
                        else if (g_ascii_strcasecmp (*buf, "[quotes]") == 0) {
                                state = PREFS_QUOTES;
                        }
                        else if (g_ascii_strcasecmp (*buf, "[templates]") == 0) {
                                state = PREFS_TEMPLATES;
                        }
                        else if (g_ascii_strcasecmp (*buf, "[current]") == 0) {
                                state = PREFS_CURRENT;
                        }
#if OLDSTYLE
                        else if (g_ascii_strcasecmp (*buf, "[execute]") == 0) {
                                PROG_WARN ("%s: sorry, [execute] is no longer supported\n",
                                           s->fname_prefs);
                                fclose (f);
                                g_free (*buf);
                                g_free (buf);
                                return FALSE;
                        }
                        else if (g_ascii_strcasecmp (*buf, "[method]") == 0) {
                                state = PREFS_METHOD;
                        }
                        else if (g_ascii_strcasecmp (*buf, "[quote file]") == 0) {
                                state = PREFS_QUOTES;
                        }
                        else if (g_ascii_strcasecmp (*buf, "[source list]") == 0) {
                                state = PREFS_TEMPLATES;
                        }
#endif
                        else if (*buf[0] == '[') {
                                PROG_WARN ("%s: unknown section: %s\n", s->fname_prefs, *buf);
                                state = PREFS_NONE;
                        }
                        else {
                                switch (state) {
                                case PREFS_TARGET:
                                        g_free (s->fname_dest);
                                        s->fname_dest = g_strdup (*buf);
                                        if (s->fname_dest == NULL) {
                                                fclose (f);
                                                PROG_WARN_MEM;
                                                exit (1);
                                        }
                                        break;
                                case PREFS_CURRENT:
                                        s->current = atoi (*buf);
                                        break;
                                case PREFS_DASH:
                                        if (g_ascii_strcasecmp (*buf, "yes") == 0
                                            || g_ascii_strcasecmp (*buf, "on") == 0) {
                                                s->dash = TRUE;
                                        }
                                        else if (g_ascii_strcasecmp (*buf, "no") == 0
                                                 || g_ascii_strcasecmp (*buf, "off") == 0) {
                                                s->dash = FALSE;
                                        }
                                        else {
                                                PROG_WARN ("%s: expected \"yes\" or \"no\" for [dash]\n",
                                                           s->fname_prefs);
                                        }
                                        break;
                                case PREFS_ROTMODE:
                                        if (g_ascii_strcasecmp (*buf, "rotate") == 0) {
                                                s->rotmode = ROTMODE_ROTATE;
                                        }
                                        else if (g_ascii_strcasecmp (*buf, "random") == 0) {
                                                s->rotmode = ROTMODE_RANDOM;
                                        }
                                        else {
                                                PROG_WARN ("%s: expected \"rotate\" or \"random\" for [mode]\n",
                                                           s->fname_prefs);
                                        }
                                        break;
#if OLDSTYLE
                                case PREFS_METHOD:
                                        if (g_ascii_strcasecmp (*buf, "link") == 0) {
                                                PROG_WARN ("%s: sorry, [method] \"link\" is no longer supported\n",
                                                           s->fname_prefs);
                                                fclose (f);
                                                g_free (*buf);
                                                g_free (buf);
                                                return FALSE;
                                        }
                                        /* The other [method] was "copy", which was the same
                                         * as the current default behaviour. So, might as well
                                         * just ignore anything other than "link".
                                         */
                                        break;
#endif
                                case PREFS_QUOTES:
                                        g_free (s->fname_quotes);
                                        s->fname_quotes = g_strdup (*buf);
                                        if (s->fname_quotes == NULL) {
                                                fclose (f);
                                                PROG_WARN_MEM;
                                                exit (1);
                                        }
                                        break;
                                case PREFS_TEMPLATES:
                                        s->template_list = g_list_append (s->template_list, g_strdup (*buf));
                                        if (s->template_list == NULL) {
                                                fclose (f);
                                                PROG_WARN_MEM;
                                                exit (1);
                                        }
                                        break;
                                case PREFS_NONE:
                                        break;
                                }
                        }
                }

        } while (len >= 0);

        fclose (f);

        g_free (*buf);
        g_free (buf);

        if (s->template_list == NULL) {
                PROG_WARN ("%s: at least one template file name is required\n", s->fname_prefs);
                return FALSE;
        }

        return TRUE;
}



static const gchar*
insert_text (Box *b, GString *gstr, const gchar *str, gchar *buf)
{
        /*
         * Copy str into gstr->str, copying a maximum of
         * b->x2 chars.
         */

        gint x, x_space, x_max, backoff, count, offset;


        if (b->strip) while (isspace (*str)) str++;
        if (*str == '\0') return str;
        if (b->x2 > gstr->len) {
                /* pad gstr->str with trailing space chars */
                gchar *padstr = g_strnfill (b->x2 - gstr->len, ' ');
                if (padstr == NULL) {
                        PROG_WARN_MEM;
                        exit (1);
                }
                g_string_append (gstr, padstr);
                if (gstr == NULL) {
                        PROG_WARN_MEM;
                        exit (1);
                }
                g_free (padstr);
        }

        x_max = b->x2 - b->x1;

        /* copy chars into buf, stripping out excess space chars */
        count = 0;
        x = 0;
        while (*str != '\0' && count <= x_max) {
                if (b->strip && isspace (*str)) {
                        buf[count] = ' ';
                        while (isspace (*str)) {
                                str++;
                        }
                        x_space = x;
                        if (*str == '\0') break;
                }
                else {
                        buf[count] = *str;
                        str++;
                }
                count++;
                x++;
        }

        /* don't split a word */
        if (b->strip && *str != '\0' && !isspace (*str)) {
                backoff = x - x_space;
                str = str - backoff;
                count = count - backoff;
        }

        /* alignment offset */
        switch (b->align) {
        case ALIGN_LEFT:
                offset = 0;
                break;
        case ALIGN_RIGHT:
                offset = 1 + x_max - count;
                break;
        case ALIGN_CENTER:
                offset = (1 + x_max - count) / 2;
                break;
        }

        /* copy buf into gstr->str */
        memcpy (&gstr->str[b->x1 - 1] + offset, buf, count);

        /* ditch trailing space chars */
        x = gstr->len;
        while (isspace (gstr->str[x - 1])) x--;
        g_string_truncate (gstr, x);


        return str;
}



static void
insert_quote (Sgr *s, Box *b, const gchar *str)
{
        /*
         * For each line in s->lines, insert str according
         * to the layout given by b->x1, b->y1, b->x2, b->y2.
         * Add extra lines to s->lines if necessary.
         */

        GList *l;
        gint i, n_lines, n_lines_needed;
        GString *gstr;
        gchar *buf;


        if (str == NULL) return;

        buf = g_malloc (b->x2 + 1);
        if (buf == NULL) {
                PROG_WARN_MEM;
                exit (1);
        }

        n_lines = g_list_length (s->lines);
        n_lines_needed = MAX (n_lines, b->y1);
        for (i = n_lines; i < n_lines_needed; i++) {
                s->lines = g_list_append (s->lines, g_string_new (""));
                if (s->lines == NULL) {
                        PROG_WARN_MEM;
                        exit (1);
                }
        }

        l = g_list_nth (s->lines, b->y1 - 1);
        while (l != NULL) {
                gstr = (GString*)l->data;
                str = insert_text (b, gstr, str, buf);
                if (b->strip) while (isspace (*str)) str++;
                if (*str != '\0' && l->next == NULL) {
                        s->lines = g_list_append (s->lines, g_string_new (""));
                        if (s->lines == NULL) {
                                PROG_WARN_MEM;
                                exit (1);
                        }
                }
                l = l->next;
        }

        g_free (buf);
}



static const gchar*
pick_quote (Box *b)
{
        /*
         * Return a pointer to one of the strings listed in
         * b->quotes, selected at random. If b->quotes is
         * NULL, return NULL.
         */

        gint n_quotes;

        if (b->quotes == NULL) return NULL;
        n_quotes = g_list_length (b->quotes);
        return (gchar*)g_list_nth (b->quotes, RANDOM_INT (n_quotes))->data;
}



static gboolean
quote_fits_box (Box *b, const gchar *str, gint len)
{
        /*
         * Return TRUE if str is going to fit the
         * requirements of Box b.
         */

        gint max_chars, x_max, y_max;
        gint n_chars, len_word, x, y;


        /* too short? */
        if (len < b->minlen) return FALSE;


        /* too long? */
        x_max = 1 + b->x2 - b->x1; /* max chars per line */
        y_max = 1 + b->y2 - b->y1; /* max number of lines */
        max_chars = x_max * y_max; /* max chars in box */

        if (!b->strip) return len <= max_chars;
        if (fast && len > max_chars) return FALSE;

        while (isspace (*str)) str++;

        x = 0;
        y = 0;
        n_chars = 0;
        while (*str != '\0' && n_chars < max_chars) {

                /* get next word length */
                len_word = 0;
                while (*str != '\0' && !isspace (*str)) {
                        str++;
                        len_word++;
                        if (len_word > x_max) return FALSE;
                }

                if (x + len_word > x_max) {
                        /* wrap */
                        y++;
                        if (y == y_max) return FALSE;
                        x = len_word;
                        if (isspace (*str)) n_chars--;
                }
                else {
                        x = x + len_word;
                }
                n_chars = n_chars + len_word;

                /* space */
                while (isspace (*str)) str++;
                x++;
                if (*str != '\0') n_chars++;

        }

        return *str == '\0' && n_chars >= b->minlen;
}



static gboolean
read_quotes (Sgr *s, Box *b)
{
        /*
         * Read the file specified by b->fname_quotes.
         * Each line in the file is a separate quote.
         * For each quote, check that it fits within
         * Box b. If it does, add it to the b->quotes
         * GList. Return FALSE if no quotes are added.
         */

        FILE    *f;
        gint    len, bufsize = 0;
        gchar   **buf;


        if (b->fname_quotes == NULL) {
                PROG_WARN ("%s: box %d: no quote file specified here or in %s\n",
                           s->fname_template, b->id, s->fname_prefs);
                return FALSE;
        }

        f = fopen (b->fname_quotes, "r");
        if (f == NULL) {
                PROG_WARN ("%s: box %d: unable to open %s\n",
                           s->fname_template, b->id, b->fname_quotes);
                return FALSE;
        }

        buf = g_malloc (sizeof (gchar*));
        if (buf == NULL) {
                fclose (f);
                PROG_WARN_MEM;
                exit (1);
        }
        *buf = NULL;

        do {
                len = getline (buf, &bufsize, f);
                if (len > 0 && keep (*buf)) {
                        if (buf[0][len - 1] == '\n') buf[0][len - 1] = '\0';
                        if (quote_fits_box (b, *buf, len - 1)) {
                                b->quotes = g_list_append (b->quotes, g_strdup (*buf));
                                if (b->quotes == NULL) {
                                        fclose (f);
                                        PROG_WARN_MEM;
                                        exit (1);
                                }
                        }
                }

        } while (len >= 0);

        fclose (f);

        g_free (*buf);
        g_free (buf);

        if (b->quotes == NULL) {
                PROG_WARN ("%s: box %d: nothing in %s fits\n",
                           s->fname_template, b->id, b->fname_quotes);
        }

        return b->quotes != NULL;
}



static Box*
parse_layout (Sgr *s, const gchar *layout, gint id, gboolean *err)
{
        /*
         * Return a new Box with values derived from the
         * layout string. The layout string is passed in
         * after the "[quote" part, so it looks something
         * like " x1=1 y1=3 x2=70 y2=3]". The format is:
         *
         * ( space+ name space* = space* ( value | ' value ' | " value ") )* ]
         *
         * where name is any string listed in attrs[].
         *
         */

        Box *b;
        AttrType attr;
        const gchar *str = layout, *beg;
        gchar *valstr;
        gint len;


        b = box_new (s, id);

        if (*str == ']') {
                /* stick with defaults */
                *err = FALSE;
                return b;
        }

        *err = TRUE; /* till success */

        if (!isspace (*str)) {
                PROG_WARN ("box %d: space expected before attribute name\n", id);
                return b;
        }

        while (*str != '\0' && *str != ']') {


                /* expect and skip space */
                str = skip_space (b, str);
                if (str == NULL) return b;

                /* expect attribute name or ']' */
                if (*str == ']') {
                        *err = FALSE;
                        return b;
                }
                attr = identify_attr (str, &len);
                if (attr == ATTR_ERROR) {
                        PROG_WARN ("box %d: unknown attribute name\n", id);
                        return b;
                }
                str = skip_chars (b, str, len);
                if (str == NULL) return b;


#if OLDSTYLE
                /* expect space or '=' */
                str = skip_space (b, str);
                if (str == NULL) return b;
                if (*str == '=') {
                        str++;
                        str = skip_space (b, str);
                        if (str == NULL) return b;
                }
#else
                /* expect '=', possibly space padded */
                str = skip_space (b, str);
                if (str == NULL) return b;
                if (*str != '=') {
                        PROG_WARN ("box %d: expected '=' after %s\n", id, attrs[attr]);
                        return b;
                }
                str++;
                str = skip_space (b, str);
                if (str == NULL) return b;
#endif

                /* expect attribute value, possibly ' or " quoted */
                if (*str == '\'' || *str == '"') {
                        gchar ch = *str;
                        str++;
                        beg = str;
                        while (*str != ch && *str != '\0') str++;
                        if (*str == '\0') {
                                PROG_WARN ("box %d: unexpected end of string\n", id);
                                return b;
                        }
                        else if (str - beg == 0) {
                                PROG_WARN ("box %d: missing value for %s\n", id, attrs[attr]);
                                return b;
                        }
                        valstr = g_strndup (beg, str - beg);
                        if (valstr == NULL) {
                                PROG_WARN_MEM;
                                exit (1);
                        }
                        str++;
                }
                else {
                        beg = str;
                        while (*str != ']' && *str != '\0' && !isspace (*str)) str++;
                        if (*str == '\0') {
                                PROG_WARN ("box %d: unexpected end of string\n", id);
                                return b;
                        }
                        else if (str - beg == 0) {
                                PROG_WARN ("box %d: missing value for %s\n", id, attrs[attr]);
                                return b;
                        }
                        valstr = g_strndup (beg, str - beg);
                        if (valstr == NULL) {
                                PROG_WARN_MEM;
                                exit (1);
                        }
                }


                /* store value according to attribute */
                switch (attr) {
                case ATTR_X1:
#if OLDSTYLE
                case ATTR_STARTX:
#endif
                        b->x1 = atoi (valstr);
                        if (b->x1 < 1) {
                                PROG_WARN ("box %d: x1 must be >= 1\n", id);
                                return b;
                        }
                        break;
                case ATTR_Y1:
#if OLDSTYLE
                case ATTR_STARTY:
#endif
                        b->y1 = atoi (valstr);
                        if (b->y1 < 1) {
                                PROG_WARN ("box %d: y1 must be >= 1\n", id);
                                return b;
                        }
                        break;
                case ATTR_X2:
#if OLDSTYLE
                case ATTR_ENDX:
#endif
                        b->x2 = atoi (valstr);
                        if (b->x2 < 1) {
                                PROG_WARN ("box %d: x2 must be >= 1\n", id);
                                return b;
                        }
                        break;
                case ATTR_Y2:
#if OLDSTYLE
                case ATTR_ENDY:
#endif
                        b->y2 = atoi (valstr);
                        if (b->y2 < 1) {
                                PROG_WARN ("box %d: y2 must be >= 1\n", id);
                                return b;
                        }
                        break;
                case ATTR_ALIGN:
                        if (g_ascii_strcasecmp (valstr, "left") == 0) {
                                b->align = ALIGN_LEFT;
                        }
                        else if (g_ascii_strcasecmp (valstr, "right") == 0) {
                                b->align = ALIGN_RIGHT;
                        }
                        else if (g_ascii_strcasecmp (valstr, "center") == 0
                                 || g_ascii_strcasecmp (valstr, "centre") == 0) {
                                b->align = ALIGN_CENTER;
                        }
                        else {
                                PROG_WARN ("box %d: align must be 'left', 'right' or 'center'\n", id);
                                return b;
                        }
                        break;
                case ATTR_MINLEN:
                        b->minlen = atoi (valstr);
                        if (b->minlen < 0) {
                                PROG_WARN ("box %d: minlen must be >= 0\n", id);
                                return b;
                        }
                        break;
                case ATTR_STRIP:
                        if (g_ascii_strcasecmp (valstr, "yes") == 0) {
                                b->strip = TRUE;
                        }
                        else if (g_ascii_strcasecmp (valstr, "no") == 0) {
                                b->strip = FALSE;
                        }
                        else {
                                PROG_WARN ("box %d: strip value must be 'yes' or 'no'\n", id);
                                return b;
                        }
                        break;
                case ATTR_FILE:
                        g_free (b->fname_quotes);
                        b->fname_quotes = expand_tilde (valstr);
                        break;
                case ATTR_ENV:
                        b->env = g_strdup (g_getenv (valstr));
                        if (b->env == NULL) {
                               PROG_WARN ("box %d: couldn't get a value for %s\n", id, valstr);
                        }
                        break;
                case ATTR_ERROR:
                        break;
                }
                g_free (valstr);


                /* expect ']' or space */
                if (*str != ']' && !isspace (*str)) {
                        PROG_WARN ("box %d: expected space or ']' after %s value\n", id, attrs[attr]);
                        return b;
                }
        }

        if (*str != ']') {
                PROG_WARN ("box %d: expected ']' at end of line\n", id);
                return b;
        }

        if (b->x2 < b->x1) {
                PROG_WARN ("box %d: x2 must be >= x1\n", id);
                return b;
        }
        if (b->y2 < b->y1) {
#if OLDSTYLE
                b->y2 = b->y1;
#else
                PROG_WARN ("box %d: y2 must be >= y1\n", id);
                return b;
#endif
        }


        *err = FALSE;

        return b;
}



static gboolean
read_template (Sgr *s)
{
        /*
         * Read the file specified by s->fname_template, storing
         * each line in the s->lines GList. Return FALSE on error.
         */

        FILE    *f;
        gint    len, bufsize = 0;
        gchar   **buf;


        f = fopen (s->fname_template, "r");
        if (f == NULL) {
                PROG_WARN ("%s: unable to read %s\n", s->fname_prefs, s->fname_template);
                return FALSE;
        }

        buf = g_malloc (sizeof (gchar*));
        if (buf == NULL) {
                fclose (f);
                PROG_WARN_MEM;
                exit (1);
        }
        *buf = NULL;

        do {
                len = getline (buf, &bufsize, f);
                if (len > 0) {
                        if (buf[0][len - 1] == '\n') buf[0][len - 1] = '\0';
                        s->lines = g_list_append (s->lines, g_string_new (*buf));
                        if (s->lines == NULL) {
                                fclose (f);
                                PROG_WARN_MEM;
                                exit (1);
                        }
                }

        } while (len >= 0);

        fclose (f);

        g_free (*buf);
        g_free (buf);

        return TRUE;
}



static gboolean
pick_template (Sgr *s)
{
        /*
         * Return FALSE if the template list is empty.
         * Update s->current according to s->rotmode.
         * Copy the currently selected template's file
         * name to s->fname_template.
         */

        gint n_sigs = g_list_length (s->template_list);

        if (n_sigs == 0) {
                PROG_WARN ("%s: no signature template files\n", s->fname_prefs);
                return FALSE;
        }

        if (s->rotmode == ROTMODE_RANDOM) {
                s->current = RANDOM_INT (n_sigs);
        }
        else {
                s->current = s->current + 1;
                if (s->current >= n_sigs) s->current = 0;
        }
        s->fname_template = expand_tilde ((gchar*)g_list_nth
                                        (s->template_list, s->current)->data);

        return TRUE;
}



static gboolean
sgrotum (Sgr *s)
{
        /*
         * Pick a signature template at random from s->template_list.
         * Read the selected template file into s->lines.
         * Insert quotes into s->lines as required.
         * Return FALSE if anything falls over.
         */

        GString *gstr_layout;
        GList *l;
        Box *b;
        gint id = 1;
        gboolean err;


        if (!pick_template (s)) return FALSE;
        if (!read_template (s)) return FALSE;

        if (s->lines != NULL) gstr_layout = (GString*)s->lines->data;
        else gstr_layout = NULL;

        /* add a Box for each "[quote ...]" in the selected template */
        while (gstr_layout != NULL && g_ascii_strncasecmp (gstr_layout->str, "[quote", 6) == 0) {
                b = parse_layout (s, &gstr_layout->str[6], id, &err);
                if (b == NULL || err) {
                        PROG_WARN ("error while working on %s\n", s->fname_template);
                        box_free (b);
                        return FALSE;
                }
                id++;

                s->box = g_list_append (s->box, b);
                if (s->box == NULL) {
                        PROG_WARN_MEM;
                        exit (1);
                }

                /* remove this [quote ...] line from s->lines */
                s->lines = g_list_remove (s->lines, gstr_layout);
                g_string_free (gstr_layout, TRUE);
                if (s->lines != NULL) {
                        gstr_layout = (GString*)s->lines->data;
                }
                else {
                        gstr_layout = NULL;
                }
        }


        /* insert a quote in each box */
        l = s->box;
        while (l != NULL) {

                b = (Box*)l->data;

                if (b->env == NULL) {
                        /* it's a normal quote */
                        if (!read_quotes (s, b)) return FALSE;
                        insert_quote (s, b, pick_quote (b));
                }
                else {
                        /* it's the value of an environment variable */
                        if (!quote_fits_box (b, b->env, strlen (b->env))) {
                                PROG_WARN ("box %d: won't fit: %s\n", b->id, b->env);
                                return FALSE;
                        }
                        insert_quote (s, b, b->env);
                }

                l = l->next;
        }

        return TRUE;
}



static void
usage (void)
{
        g_print ("\nsgrotum  - an email signature rotumbulator\n\n"
                 "Usage: " PACKAGE " [OPTIONS] preference-file\n\n"
                 "  -f, --fast      use simpler quote length testing\n"
                 "  -h, --help      display this help and exit\n"
                 "  -n, --no-write  don't write target signature file\n"
                 "  -s, --show      send resulting signature to stdout\n"
                 "  -v, --version   output version information and exit\n\n"
                 "For details: `man " PACKAGE "'\n\n");
}



static void
invokation_error (void)
{
        g_printerr ("Try `" PACKAGE " --help' for more information\n");
        exit (2);
}



int
main (int argc, char *argv[])
{
        const gchar *fname = NULL;
        Sgr *s;
        gint ch, option_index;


        while (1) {
                option_index = 0;

                ch = getopt_long (argc, argv, "fhnsv", long_options, &option_index);
                if (ch == -1) break;

                switch (ch) {
                case 'f':
                        fast = TRUE;
                        break;
                case 'h':
                        usage ();
                        exit (0);
                case 'n':
                        write = FALSE;
                        break;
                case 's':
                        show = TRUE;
                        break;
                case 'v':
#if OLDSTYLE
                        g_print (VERSION " (compiled with --enable-oldstyle)\n");
#else
                        g_print (VERSION "\n");
#endif
                        exit (0);
                default:
                        invokation_error ();
                }
        }
        if (argc == optind) {
                PROG_WARN ("no preference file specified\n");
                invokation_error ();
        }
        if (argc > optind + 1) {
                PROG_WARN ("too many arguments\n");
                invokation_error ();
        }
        fname = argv[optind];


        srand (time (NULL));


        s = sgrotum_new ();


        if (!read_prefs (s, fname)) {
                cleanup (s, 1);
        }


        if (!sgrotum (s)) {
                cleanup (s, 1);
        }


        write_signature (s);


        write_prefs (s);


        cleanup (s, 0);
        exit (0);
}
